"""Provide the abstract reduced functional, and a vanilla implementation."""
from abc import ABC, abstractmethod
from contextlib import contextmanager
from .drivers import compute_derivative, compute_hessian, compute_tlm
from .enlisting import Enlist
from .tape import get_working_tape, stop_annotating, no_annotations
from .overloaded_type import OverloadedType, create_overloaded_object
from .adjfloat import AdjFloat
from .control import Control


class AbstractReducedFunctional(ABC):
    """Base class for reduced functionals.

    An object which encompasses computations of the form::

        Jhat(m) = J(u(m), m)

    Where `u` is the system state and `m` is a `pyadjoint.Control` or list of
    `pyadjoint.Control`, `J` is an overloaded type providing the functional value,
    and `Jhat` is a reduced functional where the explicit dependence on `u` has
    been eliminated.

    A reduced functional is callable and takes as arguments the value(s) of the
    control(s) at which it is to be evaluated.

    Despite the name, the value of a reduced functional need not be scalar. If
    the functional is non-scalar valued then derivative calculations will need
    to be seeded with an input adjoint value of the adjoint type matching the
    result of the functional.
    """

    @property
    @abstractmethod
    def controls(self) -> list[Control]:
        """Return the list of controls on which this functional depends."""

    @abstractmethod
    def __call__(
        self, values: OverloadedType | list[OverloadedType]
    ) -> OverloadedType:
        """Compute the reduced functional with supplied control value.

        Args:
            values ([OverloadedType]): If you have multiple controls this
                should be a list of new values for each control in the order
                you listed the controls to the constructor. If you have a
                single control it can either be a list or a single object.
                Each new value should have the same type as the corresponding
                control.

        Returns:
            :obj:`OverloadedType`: The computed value. Often of type
                :class:`AdjFloat`.

        """
        raise NotImplementedError

    @abstractmethod
    def derivative(self, adj_input=1.0, apply_riesz=False):
        """Return the derivative of the functional w.r.t. the control.

        Using the adjoint method, the derivative of the functional with respect
        to the control, around the last supplied value of the control, is
        computed and returned.

        Args:
            adj_input: The adjoint value to the functional result. Required if
                the functional is not scalar-valued, or if the functional is an
                intermediate result in the computation of an outer functional.
            apply_riesz: If True, apply the Riesz map of each control in order
                to return a primal gradient rather than a derivative in the
                dual space.

        Returns:
            OverloadedType: The derivative with respect to the control.
                If apply_riesz is False, should be an instance of the type dual
                to that of the control. If apply_riesz is True should have the
                same type as the control.

        """
        raise NotImplementedError

    @abstractmethod
    def hessian(self, m_dot, hessian_input=None, evaluate_tlm=True, apply_riesz=False):
        """Return the action of the Hessian of the functional.

        The Hessian is evaluated w.r.t. the control on a vector m_dot.

        Using the second-order adjoint method, the action of the Hessian of the
        functional with respect to the control, around the last supplied value
        of the control, is computed and returned.

        Args:
            m_dot ([OverloadedType]): The direction in which to compute the
                action of the Hessian.
            hessian_input OverloadedType: The Hessian value for the functional result.
                Required if the functional is not scalar-valued, or if the functional
                is an intermediate result in the computation of an outer functional.
            evaluate_tlm (bool): If True, will evaluate the tangent linear model before
                evaluating the Hessian. If False, assumes that the tape is already
                populated with the tangent linear values and does not evaluate them.
            apply_riesz (bool): If True, apply the Riesz map of each control in order
                to return the (primal) Riesz representer of the Hessian
                action.

        Returns:
            OverloadedType: The action of the Hessian in the direction m_dot.
                If apply_riesz is False, should be an instance of the type dual
                to that of the control. If apply_riesz is True should have the
                same type as the control.

        """
        raise NotImplementedError

    @abstractmethod
    def tlm(self, m_dot):
        """Return the action of the tangent linear model of the functional.

        The tangent linear model is evaluated w.r.t. the control on a vector
        m_dot, around the last supplied value of the control.

        Args:
            m_dot ([OverloadedType]): The direction in which to compute the
                action of the tangent linear model.

        Returns:
            OverloadedType: The action of the tangent linear model in the direction m_dot.
                Should be an instance of the same type as the functional.
        """


def _get_extract_derivative_components(derivative_components):
    """Construct a function to pass as a pre derivative callback.

    Used when derivative components are required.
    """
    def extract_derivative_components(controls):
        controls_out = Enlist([controls[i]
                               for i in derivative_components])
        return controls_out
    return extract_derivative_components


def _get_pack_derivative_components(controls, derivative_components):
    """Construct a function to pass as a post derivative callback.

    Used when derivative components are required.
    """
    def pack_derivative_components(checkpoint, derivatives, values):
        derivatives_out = []
        count = 0
        for i, control in enumerate(controls):
            if i in derivative_components:
                derivatives_out.append(derivatives[count])
                count += 1
            else:
                zero_derivative = control._ad_copy()
                zero_derivative *= 0.
                derivatives_out.append(zero_derivative)
        return derivatives_out
    return pack_derivative_components


class ReducedFunctional(AbstractReducedFunctional):
    """Class representing the reduced functional.

    A reduced functional maps a control value to the provided functional.
    It may also be used to compute the derivative of the functional with
    respect to the control.

    Args:
        functional (:obj:`OverloadedType`): An instance of an OverloadedType,
            usually :class:`AdjFloat`. This should be the return value of the
            functional you want to reduce.
        controls (list[Control]): A list of Control instances, which you want
            to map to the functional. It is also possible to supply a single
            Control instance instead of a list.
        derivative_components (tuple of int): The indices of the controls with
            respect to which to take the derivative. By default, the derivative
            is taken with respect to all controls. If present, it overwrites
            derivative_cb_pre and derivative_cb_post.
        scale (float): A scaling factor applied to the functional and its
            gradient with respect to the control.
        tape (Tape): A tape object that the reduced functional will use to
            evaluate the functional and its gradients (or derivatives).
        eval_cb_pre (function): Callback function before evaluating the
            functional. Input is a list of Controls.
        eval_cb_pos (function): Callback function after evaluating the
            functional. Inputs are the functional value and a list of Controls.
        derivative_cb_pre (function): Callback function before evaluating
            derivatives. Input is a list of Controls.
            Should return a list of Controls (usually the same
            list as the input) to be passed to compute_derivative.
        derivative_cb_post (function): Callback function after evaluating
            derivatives.  Inputs are: functional.block_variable.checkpoint,
            list of functional derivatives, list of functional values.
            Should return a list of derivatives (usually the same
            list as the input) to be returned from self.derivative.
        hessian_cb_pre (function): Callback function before evaluating the Hessian.
            Input is a list of Controls.
        hessian_cb_post (function): Callback function after evaluating the Hessian.
            Inputs are the functional, a list of Hessian, and controls.
        tlm_cb_pre (function): Callback function before evaluating the tangent linear model.
            Input is a list of Controls.
        tlm_cb_post (function): Callback function after evaluating the tangent linear model.
            Inputs are the functional, the tlm result, and controls.
    """

    def __init__(self, functional, controls,
                 derivative_components=None,
                 scale=1.0, tape=None,
                 eval_cb_pre=lambda *args: None,
                 eval_cb_post=lambda *args: None,
                 derivative_cb_pre=lambda controls: controls,
                 derivative_cb_post=lambda checkpoint, derivative_components,
                 controls: derivative_components,
                 hessian_cb_pre=lambda *args: None,
                 hessian_cb_post=lambda *args: None,
                 tlm_cb_pre=lambda *args: None,
                 tlm_cb_post=lambda *args: None):
        if not isinstance(functional, OverloadedType):
            raise TypeError("Functional must be an OverloadedType.")
        self.functional = functional
        self.tape = get_working_tape() if tape is None else tape
        self._controls = Enlist(controls)
        self.derivative_components = derivative_components
        self.scale = scale
        self.eval_cb_pre = eval_cb_pre
        self.eval_cb_post = eval_cb_post
        self.derivative_cb_pre = derivative_cb_pre
        self.derivative_cb_post = derivative_cb_post
        self.hessian_cb_pre = hessian_cb_pre
        self.hessian_cb_post = hessian_cb_post
        self.tlm_cb_pre = tlm_cb_pre
        self.tlm_cb_post = tlm_cb_post

        if self.derivative_components:
            # pre callback
            self.derivative_cb_pre = _get_extract_derivative_components(
                derivative_components)
            # post callback
            self.derivative_cb_post = _get_pack_derivative_components(
                controls, derivative_components)

    @property
    def controls(self) -> list[Control]:
        return self._controls

    @no_annotations
    def derivative(self, adj_input=1.0, apply_riesz=False):
        values = [c.tape_value() for c in self.controls]
        controls = self.derivative_cb_pre(self.controls)

        if not controls:
            raise ValueError("""Note that the callback interface
            for derivative_cb_pre has changed. It should now return a
            list of controls (usually the same list as input.""")

        # Make sure `adj_input` is an OverloadedType
        adj_input = create_overloaded_object(adj_input)
        adj_value = adj_input._ad_mul(self.scale)

        derivatives = compute_derivative(self.functional,
                                         controls,
                                         tape=self.tape,
                                         adj_value=adj_value,
                                         apply_riesz=apply_riesz)

        # Call callback
        derivatives = self.derivative_cb_post(
            self.functional.block_variable.checkpoint,
            derivatives,
            values)

        if not derivatives:
            raise ValueError("""Note that the callback interface
            for derivative_cb_post has changed. It should now return a
            list of derivatives, usually the same list as input.""")

        return self.controls.delist(derivatives)

    @no_annotations
    def hessian(self, m_dot, hessian_input=None, evaluate_tlm=True, apply_riesz=False):
        # Call callback
        values = [c.tape_value() for c in self.controls]
        self.hessian_cb_pre(self.controls.delist(values))

        r = compute_hessian(self.functional, self.controls, m_dot,
                            hessian_input=hessian_input, tape=self.tape,
                            evaluate_tlm=evaluate_tlm, apply_riesz=apply_riesz)

        # Call callback
        self.hessian_cb_post(self.functional.block_variable.checkpoint,
                             self.controls.delist(r),
                             self.controls.delist(values))

        return self.controls.delist(r)

    @no_annotations
    def tlm(self, m_dot):
        # Call callback
        values = [c.tape_value() for c in self.controls]
        self.tlm_cb_pre(self.controls.delist(values))

        tlm = compute_tlm(self.functional, self.controls, m_dot, tape=self.tape)

        # Call callback
        self.tlm_cb_post(self.functional.block_variable.checkpoint,
                         tlm, self.controls.delist(values))

        return tlm

    @no_annotations
    def __call__(self, values):
        values = Enlist(values)
        if len(values) != len(self.controls):
            raise ValueError(
                "values should be a list of same length as controls."
            )

        for i, value in enumerate(values):
            control_type = type(self.controls[i].control)
            if isinstance(value, (int, float)) and control_type is AdjFloat:
                value = self.controls[i].control._ad_init_object(value)
            elif not isinstance(value, control_type):
                if len(values) == 1:
                    raise TypeError(
                        "Control value must be an `OverloadedType` object with the same "
                        f"type as the control, which is {control_type}"
                    )
                else:
                    raise TypeError(
                        f"The control at index {i} must be an `OverloadedType` object "
                        f"with the same type as the control, which is {control_type}"
                    )
        # Call callback.
        self.eval_cb_pre(self.controls.delist(values))

        for i, value in enumerate(values):
            self.controls[i].update(value)

        self.tape.reset_blocks()
        blocks = self.tape.get_blocks()
        self.tape._recompute_count += 1
        with self.marked_controls():
            with stop_annotating():
                if self.tape._checkpoint_manager:
                    self.tape._checkpoint_manager.recompute(self.functional)
                else:
                    for i in self.tape._bar("Evaluating functional").iter(
                        range(len(blocks))
                    ):
                        blocks[i].recompute()

        # ReducedFunctional can result in a scalar or an assembled 1-form
        func_value = self.functional.block_variable.saved_output
        # Scale the underlying functional value
        func_value *= self.scale

        # Call callback
        self.eval_cb_post(func_value, self.controls.delist(values))

        return func_value

    def optimize_tape(self):
        self.tape.optimize(
            controls=self.controls,
            functionals=[self.functional]
        )

    @contextmanager
    def marked_controls(self):
        """Return a context manager which marks the active controls."""
        for control in self.controls:
            control.mark_as_control()
        try:
            yield
        finally:
            for control in self.controls:
                control.unmark_as_control()



from .block_variable import BlockVariable
from .tape import no_annotations


class Placeholder(BlockVariable):
    """A class that functions as a placeholder for block variables (for dependencies).

    This means that you can replace the dependency of a Block with another block variable
    on the fly. Do note that Block outputs can not be placeholders.

    The placeholders are useful when you require earlier values in the computational graph
    to be values computed later in the computational graph from the previous recomputation.
    This can be the case for the initial guess for iterative solvers.
    If the recomputations are expected to only slightly change the solution,
    then using the previously found solution as initial guess can significantly speed up the recomputation.

    Usage:
        >>> u = OverloadedType()
        >>> p = Placeholder(u)
        >>> v = annotated_operator(u)
        >>> p.set_value(v)

        Each recomputation will now use the previously computed v as input
        to annotated_operator.

    """
    def __init__(self, obj):
        super(Placeholder, self).__init__(obj)
        self.block_variable = obj.block_variable
        obj.block_variable = self
        self.linked_bv = None

    def set_value(self, value):
        self.linked_bv = value.block_variable

    @no_annotations
    def save_output(self, overwrite=True):
        pass

    @property
    def saved_output(self):
        if self.linked_bv is not None:
            return self.linked_bv.saved_output
        else:
            return self.block_variable.saved_output

    def will_add_as_dependency(self):
        pass

    def will_add_as_output(self):
        pass



from enum import Enum
import sys
import gc
import logging
from functools import singledispatchmethod
from checkpoint_schedules import Copy, Move, EndForward, EndReverse, \
    Forward, Reverse, StorageType, SingleMemoryStorageSchedule
# A callback interface allowing the user to provide a
# custom error message when disk checkpointing is not configured.
disk_checkpointing_callback = {}


class CheckpointError(RuntimeError):
    pass


class Mode(Enum):
    """The mode of the checkpoint manager.

    RECORD: The forward model is being taped.
    FINISHED_RECORDING: The forward model is finished being taped.
    EVALUATED: The forward model was evaluated.
    EXHAUSTED: The forward and the adjoint models were evaluated and the schedule has concluded.
    RECOMPUTE: The forward model is being recomputed.
    EVALUATE_ADJOINT: The adjoint model is being evaluated.

    """
    RECORD = 1
    FINISHED_RECORDING = 2
    EVALUATED = 3
    EXHAUSTED = 4
    RECOMPUTE = 5
    EVALUATE_ADJOINT = 6


class CheckpointManager:
    """Manage the executions of the forward and adjoint solvers.

    Args:
        schedule (checkpoint_schedules.schedule): A schedule provided by the `checkpoint_schedules` package.
        tape (Tape): A list of blocks :class:`Block` instances.
        gc_timestep_frequency (None or int): The number of timesteps between garbage collections. The default
        is `None`, which means no invoking the garbage collector during the executions. If an integer is
        provided, the garbage collector is applied every `gc_timestep_frequency` timestep. That is useful when
        being affected by the Python fails to track and clean all checkpoint objects in memory properly.
        gc_generation (int): The generation for garbage collection. Default is 2 that runs a full collection.
        To have more information about the garbage collector generation,
        please refer to the `documentation
        <https://docs.python.org/3/library/gc.html#gc.collect>`_.

    Attributes:
        tape (Tape): A list of blocks :class:`Block` instances.
        _schedule (checkpoint_schedules.schedule): A schedule provided by the `checkpoint_schedules` package.
        forward_schedule (list): A list of `checkpoint_schedules` actions used to manage the execution of the
        forward model.
        reverse_schedule (list): A list of `checkpoint_schedules` actions used to manage the execution of the
        reverse model.
        total_timesteps (int): The total number of timesteps to execute the forward model.
        mode (Mode): The mode of the checkpoint manager. The possible modes are `RECORD`, `FINISHED_RECORDING`,
        `EVALUATED`, `EXHAUSTED`, `RECOMPUTE`, and `EVALUATE_ADJOINT`. Additional information about the modes
        can be found class:`Mode`.
        _current_action (checkpoint_schedules.CheckpointAction): The current `checkpoint_schedules` action.

    """
    def __init__(self, schedule, tape, gc_timestep_frequency=None, gc_generation=2):
        if (
            schedule.uses_storage_type(StorageType.DISK)
            and not tape._package_data
        ):
            raise CheckpointError(
                "The schedule employs disk checkpointing but it is not configured.\n"
                + "\n".join(disk_checkpointing_callback.values()))
        self.tape = tape
        self._schedule = schedule
        self.forward_schedule = []
        self.reverse_schedule = []
        if self._schedule.max_n:
            self.total_timesteps = schedule.max_n
        else:
            # We have schedules in `checkpoint_schedules` offering the flexibility to determine
            # the desired steps during the forward execution. For this type of schedule, we do not
            # have the total number of timesteps `self._schedule.max_n`. Therefore, we set the
            # `self.total_timesteps` to the maximum value of the `sys.maxsize` to indicate that
            # the total number of timesteps is not known.
            self.total_timesteps = sys.maxsize
        self.mode = Mode.RECORD
        self._current_action = next(self._schedule)
        self.forward_schedule.append(self._current_action)
        # Tell the tape to only checkpoint input data until told otherwise.
        self.tape.latest_checkpoint = 0
        self._keep_init_state_in_work = False
        self._gc_timestep_frequency = gc_timestep_frequency
        self._gc_generation = gc_generation
        # ``self._global_deps`` stores checkpoint dependencies that remain unchanged across
        # timesteps (``self.tape.timesteps``). During the forward taping process, the code
        # checks whether a dependency is in ``self._global_deps`` to avoid unnecessary clearing
        # and recreation of its checkpoint data.
        self._global_deps = set()
        self.end_timestep(-1)

    def end_timestep(self, timestep):
        """Mark the end of one timestep when taping the forward model.

        Args:
            timestep (int): The current timestep.
        """
        if self.mode == Mode.EVALUATED:
            raise CheckpointError("Not enough timesteps in schedule.")
        elif self.mode != Mode.RECORD:
            raise CheckpointError(f"Cannot end timestep in {self.mode}")
        while not self.process_taping(self._current_action, timestep + 1):
            self._current_action = next(self._schedule)
            self.forward_schedule.append(self._current_action)

    def end_taping(self):
        """Process the end of the forward execution."""
        current_timestep = self.tape.latest_timestep
        if not self._schedule.max_n:
            # Inform the schedule that the forward model has finished.
            self._schedule.finalize(len(self.tape.timesteps))
            # `self._schedule.finalize` updates `self._schedule.max_n`.
            self.total_timesteps = self._schedule.max_n
        while self.mode != Mode.EVALUATED:
            self.end_timestep(current_timestep)
            current_timestep += 1

    @singledispatchmethod
    def process_taping(self, cp_action, timestep):
        """Implement checkpointing schedule actions while taping.

        A single-dispatch generic function.

        Note:
            To have more information about the `checkpoint_schedules`, please refer to the
            `documentation <https://www.firedrakeproject.org/checkpoint_schedules/>`_.
            Detailed descriptions of the actions used in the process taping can be found at the following links:
            `Forward <https://www.firedrakeproject.org/checkpoint_schedules/checkpoint_schedules.html#
            checkpoint_schedules.schedule.Forward>`_ and `End_Forward <https://www.firedrakeproject.org/
            checkpoint_schedules/checkpoint_schedules.html#checkpoint_schedules.schedule.EndForward>`_.

        Args:
            cp_action (checkpoint_schedules.CheckpointAction): A checkpoint action obtained from the
            `checkpoint_schedules`.
            timestep (int): The current timestep.

        Returns:
            bool: Returns `True` if the timestep is in the `checkpoint_schedules` action.
            For example, if the `checkpoint_schedules` action is `Forward(0, 4, True, False, StorageType.DISK)`,
            then timestep `0, 1, 2, 3` is considered within the action; timestep `4` is not considered within the
            action and `False` is returned.

        Raises:
            CheckpointError: If the checkpoint action is not supported.
        """

        raise CheckpointError(f"Unable to process {cp_action} while taping.")

    @process_taping.register(Forward)
    def _(self, cp_action, timestep):
        if timestep < (cp_action.n0):
            raise CheckpointError(
                "Timestep is before start of Forward action."
            )

        self.tape._eagerly_checkpoint_outputs = cp_action.write_adj_deps
        _store_checkpointable_state = False
        _store_adj_dependencies = False
        if timestep > cp_action.n0 and cp_action.storage != StorageType.WORK:
            if cp_action.write_ics and timestep == (cp_action.n0 + 1):
                # Store the checkpoint data. This is the required data for
                # restarting the forward model from the step `n0`.
                _store_checkpointable_state = True
            if cp_action.write_adj_deps:
                # Store the checkpoint data. This is the required data for
                # computing the adjoint model from the step `n1`.
                _store_adj_dependencies = True
            if (
                (_store_checkpointable_state or _store_adj_dependencies)
                and cp_action.storage == StorageType.DISK
            ):
                for package in self.tape._package_data.values():
                    package.continue_checkpointing()
            if timestep == 1:
                # Store the possible global dependencies.
                for deps in self.tape.timesteps[timestep - 1].checkpointable_state:
                    self._global_deps.add(deps)
            else:
                # Check if the block variables stored in `self._global_deps` are still
                # dependencies in the previous timestep. If not, remove them from the
                # global dependencies.
                deps_to_clear = self._global_deps.difference(
                    self.tape.timesteps[timestep - 1].checkpointable_state)

                # Remove the block variables that are not global dependencies.
                self._global_deps.difference_update(deps_to_clear)

                # For no global dependencies, checkpoint storage occurs at a self.tape
                # timestep only when required by an action from the schedule. Thus, we
                # have to clear the checkpoint of block variables excluded from the self._global_deps.
                for deps in deps_to_clear:
                    deps._checkpoint = None

            self.tape.timesteps[timestep - 1].checkpoint(
                _store_checkpointable_state, _store_adj_dependencies, self._global_deps)
            # Remove unnecessary variables in working memory from previous steps.
            for var in self.tape.timesteps[timestep - 1].checkpointable_state - self._global_deps:
                var._checkpoint = None

            for block in self.tape.timesteps[timestep - 1]:
                for out in block.get_outputs():
                    out._checkpoint = None

        if cp_action.storage == StorageType.DISK:
            # Activate disk checkpointing only in the checkpointing process.
            for package in self.tape._package_data.values():
                package.pause_checkpointing()

        if isinstance(self._gc_timestep_frequency, int) and timestep % self._gc_timestep_frequency == 0:
            logging.info("Running a garbage collection cycle")
            gc.collect(self._gc_generation)

        if timestep in cp_action and timestep < self.total_timesteps:
            self.tape.get_blocks().append_step()
            if cp_action.write_ics:
                self.tape.latest_checkpoint = cp_action.n0
            return True
        else:
            return False

    @process_taping.register(EndForward)
    def _(self, cp_action, timestep):
        if timestep != self.total_timesteps:
            raise CheckpointError(
                "The correct number of forward steps has notbeen taken."
            )
        self.mode = Mode.EVALUATED
        return True

    def recompute(self, functional=None):
        """Recompute the forward model.

        Args:
            functional (BlockVariable): The functional to be evaluated.
        """
        if self.mode == Mode.RECORD:
            # Finalise the taping process.
            self.end_taping()
        if self._schedule.uses_storage_type(StorageType.DISK):
            # Clear the data of the current state before recomputing.
            for package in self.tape._package_data.values():
                package.reset()
        self.mode = Mode.RECOMPUTE
        with self.tape.progress_bar("Evaluating Functional", max=self.total_timesteps) as progress_bar:
            # Restore the initial condition to advance the forward model from the step 0.
            current_step = self.tape.timesteps[self.forward_schedule[0].n0]
            current_step.restore_from_checkpoint(self.forward_schedule[0].storage)
            for cp_action in self.forward_schedule:
                self._current_action = cp_action
                self.process_operation(cp_action, progress_bar, functional=functional)

    def evaluate_adj(self, last_block, markings):
        """Evaluate the adjoint model.

        Args:
            last_block (int): The last block to be evaluated.
            markings (bool): If `True`, then each `BlockVariable` of the current block will have set
            `is_control_dependent` attribute indicating whether their adjoint components are relevant
            for computing the final target adjoint values and `is_functional_dependency` indicating
            whether the functional depends on their value.
        """
        # Work out other cases when they arise.
        if last_block != 0:
            raise NotImplementedError(
                "Only the first block can be evaluated at present."
            )
        if self.mode == Mode.RECORD:
            # Finalise the taping process.
            self.end_taping()

        if self.mode not in (Mode.EVALUATED, Mode.FINISHED_RECORDING):
            raise CheckpointError("Evaluate Functional before calling gradient.")

        with self.tape.progress_bar("Evaluating Adjoint", max=self.total_timesteps) as progress_bar:
            if self.reverse_schedule:
                for cp_action in self.reverse_schedule:
                    self.process_operation(cp_action, progress_bar, markings=markings)
            else:
                while not isinstance(self._current_action, EndReverse):
                    cp_action = next(self._schedule)
                    self._current_action = cp_action
                    self.reverse_schedule.append(cp_action)
                    self.process_operation(cp_action, progress_bar, markings=markings)

            # Only set the mode after the first backward in order to handle
            # that step correctly.
            self.mode = Mode.EVALUATE_ADJOINT

    @singledispatchmethod
    def process_operation(self, cp_action, progress_bar, **kwargs):
        """A function used to process the forward and adjoint executions.
        This single-dispatch generic function is used in the `Blocks`
        recomputation and adjoint evaluation with checkpointing.

        Note:
            The documentation of the `checkpoint_schedules` actions is available
            `here <https://www.firedrakeproject.org/checkpoint_schedules/>`_.

        Args:
            cp_action (checkpoint_schedules.CheckpointAction): A checkpoint action obtained from the
            `checkpoint_schedules`.
            progress_bar (progressbar.ProgressBar): A progress bar to display the progress of the reverse executions.
            kwargs: Additional keyword arguments.

        Raises:
            CheckpointError: If the checkpoint action is not supported.
        """
        raise CheckpointError(f"Unable to process {cp_action}.")

    @process_operation.register(Forward)
    def _(self, cp_action, progress_bar, functional=None, **kwargs):
        step = cp_action.n0
        #  In a dynamic schedule `cp_action` can be unbounded so we also need
        # to check `self.total_timesteps`.
        while step in cp_action and step < self.total_timesteps:
            if self.mode == Mode.RECOMPUTE and progress_bar:
                progress_bar.next()
            # Get the blocks of the current step.
            current_step = self.tape.timesteps[step]
            for block in current_step:
                block.recompute()
            _store_checkpointable_state = False
            _store_adj_dependencies = False
            if cp_action.storage != StorageType.WORK:
                if (cp_action.write_ics and step == cp_action.n0):
                    _store_checkpointable_state = True
                if cp_action.write_adj_deps:
                    _store_adj_dependencies = True
                if (
                    (_store_checkpointable_state or _store_adj_dependencies)
                    and cp_action.storage == StorageType.DISK
                ):
                    for package in self.tape._package_data.values():
                        package.continue_checkpointing()
                current_step.checkpoint(
                    _store_checkpointable_state, _store_adj_dependencies, self._global_deps)

            to_keep = set()
            if step < (self.total_timesteps - 1):
                next_step = self.tape.timesteps[step + 1]
                # The checkpointable state set of the current step.
                to_keep = next_step.checkpointable_state
            if functional:
                to_keep = to_keep.union([functional.block_variable])

            for var in current_step.checkpointable_state - to_keep.union(self._global_deps):
                # Handle the case where step is 0
                if step == 0 and var not in current_step._checkpoint:
                    # Ensure initialisation state is kept.
                    self._keep_init_state_in_work = True
                    break

                # Handle the case for SingleMemoryStorageSchedule
                if isinstance(self._schedule, SingleMemoryStorageSchedule):
                    if step > 1 and var not in self.tape.timesteps[step - 1].adjoint_dependencies:
                        var._checkpoint = None
                    continue

                # Handle variables in the initial timestep
                if (
                    var in self.tape.timesteps[0].checkpointable_state
                    and self._keep_init_state_in_work
                ):
                    continue

                # Clear the checkpoint for other cases
                var._checkpoint = None

            for block in current_step:
                # Remove unnecessary variables from previous steps.
                for bv in block.get_outputs():
                    if (
                        (cp_action.write_adj_deps and cp_action.storage != StorageType.WORK)
                        or not cp_action.write_adj_deps
                    ):
                        if bv not in to_keep:
                            bv._checkpoint = None
                    else:
                        if bv not in current_step.adjoint_dependencies.union(to_keep):
                            bv._checkpoint = None

            if self._gc_timestep_frequency and step % self._gc_timestep_frequency == 0:
                logging.info("Running a garbage collection cycle")
                gc.collect(self._gc_generation)

            step += 1
            if cp_action.storage == StorageType.DISK:
                # Activate disk checkpointing only in the checkpointing process.
                for package in self.tape._package_data.values():
                    package.pause_checkpointing()

    @process_operation.register(Reverse)
    def _(self, cp_action, progress_bar, markings, functional=None, **kwargs):
        for step in cp_action:
            if progress_bar:
                progress_bar.next()
            # Get the blocks of the current step.
            current_step = self.tape.timesteps[step]
            for block in reversed(current_step):
                block.evaluate_adj(markings=markings)
                if not current_step._revised_adj_deps:
                    # Update the adjoint dependency set.
                    for deps in block.get_dependencies():
                        if deps.is_functional_dependency:
                            current_step.adjoint_dependencies.add(deps)
                    for out in block._outputs:
                        if not out.is_functional_dependency:
                            current_step.adjoint_dependencies.discard(out)
            current_step._revised_adj_deps = True
            # Output variables are used for the last time when running
            # backwards.
            to_keep = current_step.checkpointable_state
            if functional:
                to_keep = to_keep.union([functional.block_variable])
            for block in current_step:
                block.reset_adjoint_state()
                for out in block.get_outputs():
                    out.reset_variables(("tlm",))
                    if not out.is_control:
                        out.reset_variables(("adjoint", "hessian"))
                    if cp_action.clear_adj_deps and out not in to_keep:
                        out._checkpoint = None
            if self._gc_timestep_frequency and step % self._gc_timestep_frequency == 0:
                logging.info("Running a garbage collection cycle")
                gc.collect(self._gc_generation)

    @process_operation.register(Copy)
    def _(self, cp_action, progress_bar, **kwargs):
        current_step = self.tape.timesteps[cp_action.n]
        current_step.restore_from_checkpoint(cp_action.from_storage)

    @process_operation.register(Move)
    def _(self, cp_action, progress_bar, **kwargs):
        current_step = self.tape.timesteps[cp_action.n]
        current_step.restore_from_checkpoint(cp_action.from_storage)
        current_step.delete_checkpoint()

    @process_operation.register(EndForward)
    def _(self, cp_action, progress_bar, **kwargs):
        self.mode = Mode.EVALUATED

    @process_operation.register(EndReverse)
    def _(self, cp_action, progress_bar, **kwargs):
        if self._schedule.is_exhausted:
            self.mode = Mode.EXHAUSTED
        else:
            self.mode = Mode.EVALUATED



from __future__ import print_function

from ..tape import no_annotations

# TODO: There might be a better way to handle this.
try:
    import moola

    _moola_installed = True
except ImportError:
    _moola_installed = False

__all__ = ["MoolaOptimizationProblem"]


def MoolaOptimizationProblem(rf, memoize=1):
    """Build the moola problem from the OptimizationProblem instance.
       memoize describes the number of the function and derivative
       calls to be memoized
    """

    try:
        import moola
    except ImportError:
        print("You need to install moola. Try `pip install moola`")
        raise

    class Functional(moola.Functional):
        latest_eval_hash = []
        latest_eval_eval = []
        latest_deriv_hash = []
        latest_deriv_eval = []

        @no_annotations
        def __call__(self, x):
            """ Evaluates the functional for the given control value. """
            if memoize > 0:
                hashx = hash(x)

                for hashp in self.latest_eval_hash:
                    if hashp == hashx:
                        moola.events.increment("Cached functional evaluation")
                        return self.latest_eval_eval[self.latest_eval_hash.index(hashp)]

                if len(self.latest_eval_hash) == memoize:
                    self.latest_eval_hash.pop(0)
                    self.latest_eval_eval.pop(0)

                self.latest_eval_hash.append(hashx)
                self.latest_eval_eval.append(rf(x.data))
                moola.events.increment("Functional evaluation")
                return self.latest_eval_eval[-1]

            else:
                moola.events.increment("Functional evaluation")
                return rf(x.data)

        @no_annotations
        def derivative(self, x):
            """ Evaluates the gradient for the control values. """

            if memoize > 0:
                hashx = hash(x)

                for hashp in self.latest_deriv_hash:
                    if hashp == hashx:
                        moola.events.increment("Cached derivative evaluation")
                        deriv = self.latest_deriv_eval[self.latest_deriv_hash.index(hashp)]
                        return deriv

                if len(self.latest_deriv_hash) == memoize:
                    self.latest_deriv_hash.pop(0)
                    self.latest_deriv_eval.pop(0)

                moola.events.increment("Derivative evaluation")
                D = rf.derivative()

                deriv = moola.convert_to_moola_dual_vector(D, x)

                self.latest_deriv_hash.append(hashx)
                self.latest_deriv_eval.append(deriv)
                return deriv

            else:
                moola.events.increment("Derivative evaluation")
                D = rf.derivative()

                deriv = moola.convert_to_moola_dual_vector(D, x)

                return deriv

        @no_annotations
        def hessian(self, x):
            """ Evaluates the gradient for the control values. """

            self(x)

            @no_annotations
            def moola_hessian(direction):
                moola.events.increment("Hessian evaluation")
                hes = rf.hessian(direction.data)
                return moola.convert_to_moola_dual_vector(hes, x)

            return moola_hessian

    functional = Functional()
    return moola.Problem(functional)


if _moola_installed:
    # Wrap all moola solve routines in no annotations
    moola.NewtonCG.solve = no_annotations(moola.NewtonCG.solve)
    moola.BFGS.solve = no_annotations(moola.BFGS.solve)
    moola.HybridCG.solve = no_annotations(moola.HybridCG.solve)
    moola.TrustRegionNewtonCG.solve = no_annotations(moola.TrustRegionNewtonCG.solve)
    moola.NonLinearCG.solve = no_annotations(moola.NonLinearCG.solve)
    moola.SteepestDescent.solve = no_annotations(moola.SteepestDescent.solve)



from __future__ import print_function

import numpy

from . import constraints
from .optimization_problem import MaximizationProblem
from .optimization_solver import OptimizationSolver
from ..reduced_functional_numpy import ReducedFunctionalNumPy
from ..reduced_functional_numpy import gather


class _IPOptProblem:
    """API used by cyipopt for wrapping the problem"""
    def __init__(self, objective, gradient, constraints, jacobian):
        self.objective = objective
        self.gradient = gradient
        self.constraints = constraints
        self.jacobian = jacobian


class IPOPTSolver(OptimizationSolver):
    """Use the cyipopt bindings to IPOPT to solve the given optimization problem.

    The cyipopt Problem instance is accessible as solver.ipopt_problem."""

    def __init__(self, problem, parameters=None):
        OptimizationSolver.__init__(self, problem, parameters)

        self.__build_ipopt_problem()
        self.__set_parameters()

    def __build_ipopt_problem(self):
        """Build the ipopt problem from the OptimizationProblem instance."""

        from pyadjoint.ipopt import cyipopt

        self.rfn = ReducedFunctionalNumPy(self.problem.reduced_functional)

        (lb, ub) = self.__get_bounds()
        (nconstraints, fun_g, jac_g, clb, cub) = self.__get_constraints()

        # A callback that evaluates the functional and derivative.
        J = self.rfn.__call__
        dJ = self.rfn.derivative
        nlp = cyipopt.Problem(
            n=len(ub),  # length of control vector
            lb=lb,  # lower bounds on control vector
            ub=ub,  # upper bounds on control vector
            m=nconstraints,  # number of constraints
            cl=clb,  # lower bounds on constraints
            cu=cub,  # upper bounds on constraints
            problem_obj=_IPOptProblem(
                objective=J,  # to evaluate the functional
                gradient=dJ,  # to evaluate the gradient
                constraints=fun_g,  # to evaluate the constraints
                jacobian=jac_g,  # to evaluate the constraint Jacobian
            ),
        )

        """
        if rank(self.problem.reduced_functional.mpi_comm()) > 0:
            nlp.addOption('print_level', 0)    # disable redundant IPOPT output in parallel
        else:
            nlp.addOption('print_level', 6)    # very useful IPOPT output
        """
        # TODO: Earlier the commented out code above was present.
        # Figure out how to solve parallel output cases like these in pyadjoint.
        nlp.add_option("print_level", 6)

        if isinstance(self.problem, MaximizationProblem):
            # multiply objective function by -1 internally in
            # ipopt to maximise instead of minimise
            nlp.add_option('obj_scaling_factor', -1.0)

        self.ipopt_problem = nlp

    def __get_bounds(self):
        r"""Convert the bounds into the format accepted by ipopt (two numpy arrays,
        one for the lower bound and one for the upper).

        FIXME: Do we really have to pass (-\infty, +\infty) when there are no bounds?"""

        bounds = self.problem.bounds

        if bounds is not None:
            lb_list = []
            ub_list = []  # a list of numpy arrays, one for each control

            for (bound, control) in zip(bounds, self.rfn.controls):
                general_lb, general_ub = bound  # could be float, Constant, or Function

                if isinstance(general_lb, (float, int)):
                    len_control = len(self.rfn.get_global(control))
                    lb = numpy.array([float(general_lb)] * len_control)
                else:
                    lb = self.rfn.get_global(general_lb)

                lb_list.append(lb)

                if isinstance(general_ub, (float, int)):
                    len_control = len(self.rfn.get_global(control))
                    ub = numpy.array([float(general_ub)] * len_control)
                else:
                    ub = self.rfn.get_global(general_ub)

                ub_list.append(ub)

            ub = numpy.concatenate(ub_list)
            lb = numpy.concatenate(lb_list)

        else:
            # Unfortunately you really need to specify bounds, I think?!
            ncontrols = len(self.rfn.get_controls())
            max_float = numpy.finfo(numpy.double).max
            ub = numpy.array([max_float] * ncontrols)

            min_float = numpy.finfo(numpy.double).min
            lb = numpy.array([min_float] * ncontrols)

        return (lb, ub)

    def __get_constraints(self):
        constraint = self.problem.constraints

        if constraint is None:
            # The length of the constraint vector
            nconstraints = 0

            # The bounds for the constraint
            empty = numpy.array([], dtype=float)
            clb = empty
            cub = empty

            # The constraint function, should do nothing
            def fun_g(x, user_data=None):
                return empty

            # The constraint Jacobian
            def jac_g(x, user_data=None):
                return empty

            return (nconstraints, fun_g, jac_g, clb, cub)

        else:
            # The length of the constraint vector
            nconstraints = constraint._get_constraint_dim()
            # ncontrols = len(self.rfn.get_controls())

            # The constraint function
            def fun_g(x, user_data=None):
                out = numpy.array(constraint.function(x), dtype=float)
                return out

            # The constraint Jacobian:
            # flag = True  means 'tell me the sparsity pattern';
            # flag = False means 'give me the damn Jacobian'.
            def jac_g(x, user_data=None):
                j = constraint.jacobian(x)
                out = numpy.array(gather(j), dtype=float)
                return out

            # The bounds for the constraint: by the definition of our
            # constraint type, the lower bound is always zero,
            # whereas the upper bound is either zero or infinity,
            # depending on whether it's an equality constraint or inequalityconstraint.

            clb = numpy.array([0] * nconstraints)

            def constraint_ub(c):
                if isinstance(c, constraints.EqualityConstraint):
                    return [0] * c._get_constraint_dim()
                elif isinstance(c, constraints.InequalityConstraint):
                    return [numpy.inf] * c._get_constraint_dim()

            cub = numpy.array(sum([constraint_ub(c) for c in constraint], []))

            return (nconstraints, fun_g, jac_g, clb, cub)

    _param_map = {
        'tolerance': 'tol',
        'maximum_iterations': 'max_iter',
    }

    def __set_parameters(self):
        """Set some basic parameters from the parameters dictionary that the user
        passed in, if any."""

        if self.parameters is not None:
            for param, value in self.parameters.items():
                # some parameters have a different name in ipopt
                param = self._param_map.get(param, param)
                self.ipopt_problem.add_option(param, value)

    def solve(self):
        """Solve the optimization problem and return the optimized controls."""
        guess = self.rfn.get_controls()
        results = self.ipopt_problem.solve(guess)
        new_params = [control.copy_data() for control in self.rfn.controls]
        self.rfn.set_local(new_params, results[0])

        return self.rfn.controls.delist(new_params)



"""This module offers a standard interface for control constraints,
that can be used with different optimisation algorithms."""

import copy

import numpy


class Constraint(object):
    def function(self, m):
        """
        Evaluate c(m), where c(m) == 0 for equality constraints and c(m) >= 0 for inequality constraints.

        c(m) must return a numpy array or a dolfin Function or Constant.
        """

        raise NotImplementedError("Constraint.function must be supplied")

    def jacobian(self, m):
        """Returns the full Jacobian matrix as a list of vector-like objects representing the gradient of
        the constraint function with respect to the parameter m.

        The objects returned must be of the same type as m's data."""

        raise NotImplementedError("Constraint.jacobian not implemented")

    def jacobian_action(self, m, dm, result):
        """Computes the Jacobian action of c(m) in direction dm and stores the result in result. """

        raise NotImplementedError("Constraint.jacobian_action is not implemented")

    def jacobian_adjoint_action(self, m, dp, result):
        """Computes the Jacobian adjoint action of c(m) in direction dp and stores the result in result. """

        raise NotImplementedError("Constraint.jacobian_adjoint_action is not implemented")

    def hessian_action(self, m, dm, dp, result):
        """Computes the Hessian action of c(m) in direction dm and dp and stores the result in result. """

        raise NotImplementedError("Constraint.hessian_action is not implemented")

    def output_workspace(self):
        """Return an object like the output of c(m) for calculations."""

        raise NotImplementedError("Constraint.output_workspace must be supplied")

    def _get_constraint_dim(self):
        """Returns the number of constraint components."""
        workspace = self.output_workspace()

        if hasattr(workspace, "_ad_dim"):
            return workspace._ad_dim()
        return len(workspace)


class EqualityConstraint(Constraint):
    """This class represents equality constraints of the form

    c_i(m) == 0

    for 0 <= i < n, where m is the parameter.
    """


class InequalityConstraint(Constraint):
    """This class represents constraints of the form

    c_i(m) >= 0

    for 0 <= i < n, where m is the parameter.
    """


numpify = lambda x: numpy.array(x) if isinstance(x, list) else x


class MergedConstraints(Constraint):
    def __init__(self, constraints):
        self.constraints = constraints

    def function(self, m):
        return [numpify(c.function(m)) for c in self.constraints]

    def jacobian(self, m):
        return [c.jacobian(m) for c in self.constraints]

    def jacobian_action(self, m, dm, result):
        [c.jacobian_action(m, dm, result[i]) for (i, c) in enumerate(self.constraints)]

    def jacobian_adjoint_action(self, m, dp, result):
        result._ad_imul(0.0)
        tmp = copy.deepcopy(result)

        for (i, c) in enumerate(self.constraints):
            c.jacobian_adjoint_action(m, dp[i], tmp)
            result._ad_iadd(tmp)

    def hessian_action(self, m, dm, dp, result):
        result._ad_imul(0.0)
        tmp = copy.deepcopy(result)

        for (i, c) in enumerate(self.constraints):
            c.hessian_action(m, dm, dp[i], tmp)
            result._ad_iadd(tmp)

    def __iter__(self):
        return iter(self.constraints)

    def output_workspace(self):
        return [numpify(c.output_workspace()) for c in self.constraints]

    def equality_constraints(self):
        """ Filters out the equality constraints """
        constraints = [c for c in self.constraints if isinstance(c, EqualityConstraint)]
        return MergedConstraints(constraints)

    def inequality_constraints(self):
        """ Filters out the inequality constraints """
        constraints = [c for c in self.constraints if isinstance(c, InequalityConstraint)]
        return MergedConstraints(constraints)

    def _get_constraint_dim(self):
        """ Returns the number of constraint components """
        return sum([c._get_constraint_dim() for c in self.constraints])


def canonicalise(constraints):
    if constraints is None:
        return None

    if isinstance(constraints, MergedConstraints):
        return constraints

    if not isinstance(constraints, list):
        return MergedConstraints([constraints])

    else:
        return MergedConstraints(constraints)



from .tape import annotate_tape, get_working_tape, stop_annotating
from .overloaded_type import create_overloaded_object
from .enlisting import Enlist


def overload_function(func, block_class):
    """Create an overloaded version of a function.

    This method makes several assumptions:
    1) The function is explicit, i.e. y = func(x), where y is the output of the operation.
    2) All of y is possible to convert to an OverloadedType.
    3) Unless annotation is turned off, the operation should always be annotated when calling the overloaded function.

    Args:
        func (function): The target function for which to create an overloaded version.
        block_class (type): The Block-subclass that corresponds to `func`.

    Returns:
        function: An overloaded version of `func`

    """
    def _overloaded_function(*args, **kwargs):
        annotate = annotate_tape(kwargs)
        if annotate:
            b_kwargs = block_class.pop_kwargs(kwargs)
            b_kwargs.update(kwargs)
            block = block_class(*args, **b_kwargs)

        with stop_annotating():
            func_output = func(*args, **kwargs)

        r = []
        func_output = Enlist(func_output)
        for out in func_output:
            r.append(create_overloaded_object(out))

        if annotate:
            for out in r:
                block.add_output(out.create_block_variable())
            tape = get_working_tape()
            tape.add_block(block)

        return func_output.delist(r)
    return _overloaded_function


def overloaded_function(block_class):
    """Returns a decorator for functions that should be overloaded

    Args:
        block_class (type): The Block-subclass that corresponds to the function being decorated.

    """
    def decorator(func):
        return overload_function(func, block_class)
    return decorator



from .tape import no_annotations, get_working_tape


class BlockVariable(object):
    """References a block output variable.

    """

    def __init__(self, output):
        self.output = output
        self.adj_value = None
        self.tlm_value = None
        self.hessian_value = None
        self._checkpoint = None
        self.is_control = False
        self.floating_type = False
        # Helper flags for use during tape traversals.
        self.is_control_dependent = False
        self.is_functional_dependency = False
        # By default assume the variable is created externally to the tape.
        self.creation_timestep = -1
        # The timestep during which this variable was last used as an input.
        self.last_use = -1

    def add_adj_output(self, val):
        if self.adj_value is None:
            self.adj_value = val
        else:
            self.adj_value += val

    def add_tlm_output(self, val):
        if self.tlm_value is None:
            self.tlm_value = val
        else:
            self.tlm_value += val

    def add_hessian_output(self, val):
        if self.hessian_value is None:
            self.hessian_value = val
        else:
            self.hessian_value += val

    def reset_variables(self, types):
        if "adjoint" in types:
            self.adj_value = None

        if "hessian" in types:
            self.hessian_value = None

        if "tlm" in types:
            self.tlm_value = None

    @no_annotations
    def save_output(self, overwrite=True):
        if overwrite or self.checkpoint is None:
            self._checkpoint = self.output._ad_create_checkpoint()

    @property
    def saved_output(self):
        if self.checkpoint is not None:
            return self.output._ad_restore_at_checkpoint(self.checkpoint)
        else:
            return self.output

    def will_add_as_dependency(self):
        overwrite = self.output._ad_will_add_as_dependency()
        overwrite = bool(overwrite)
        tape = get_working_tape()
        if self.last_use < tape.latest_checkpoint:
            self.save_output(overwrite=overwrite)
        tape.add_to_checkpointable_state(self, self.last_use)
        self.last_use = tape.latest_timestep

    def will_add_as_output(self):
        tape = get_working_tape()
        self.creation_timestep = tape.latest_timestep
        self.last_use = self.creation_timestep
        overwrite = self.output._ad_will_add_as_output()
        overwrite = bool(overwrite)
        if not overwrite:
            self._checkpoint = None
        if tape._eagerly_checkpoint_outputs:
            self.save_output()
        tape.add_to_adjoint_dependencies(self, self.last_use - 1)

    def __str__(self):
        return str(self.output._ad_str)

    @property
    def checkpoint(self):
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, value):
        if self.is_control:
            return
        self._checkpoint = value
